using System.Linq;
using Xunit;

namespace Kiota.Builder.Refiners.Tests {
    public class JavaLanguageRefinerTests {
        private readonly CodeNamespace root = CodeNamespace.InitRootNamespace();
        #region CommonLanguageRefinerTests
        [Fact]
        public void EscapesReservedKeywordsInInternalDeclaration() {
            var model = root.AddClass(new CodeClass {
                Name = "break",
                ClassKind = CodeClassKind.Model
            }).First();
            var nUsing = new CodeUsing {
                Name = "some.ns",
            };
            nUsing.Declaration = new CodeType {
                Name = "break",
                IsExternal = false,
            };
            model.AddUsing(nUsing);
            ILanguageRefiner.Refine(new GenerationConfiguration { Language = GenerationLanguage.Java }, root);
            Assert.NotEqual("break", nUsing.Declaration.Name);
            Assert.Contains("escaped", nUsing.Declaration.Name);
        }
        [Fact]
        public void EscapesReservedKeywords() {
            var model = root.AddClass(new CodeClass {
                Name = "break",
                ClassKind = CodeClassKind.Model
            }).First();
            ILanguageRefiner.Refine(new GenerationConfiguration { Language = GenerationLanguage.Java }, root);
            Assert.NotEqual("break", model.Name);
            Assert.Contains("escaped", model.Name);
        }
        [Fact]
        public void AddsDefaultImports() {
            var model = root.AddClass(new CodeClass {
                Name = "model",
                ClassKind = CodeClassKind.Model
            }).First();
            var requestBuilder = root.AddClass(new CodeClass {
                Name = "rb",
                ClassKind = CodeClassKind.RequestBuilder,
            }).First();
            requestBuilder.AddMethod(new CodeMethod {
                Name = "get",
                MethodKind = CodeMethodKind.RequestExecutor,
            });
            ILanguageRefiner.Refine(new GenerationConfiguration { Language = GenerationLanguage.Java }, root);
            Assert.NotEmpty(model.StartBlock.Usings);
            Assert.NotEmpty(requestBuilder.StartBlock.Usings);
        }
        [Fact]
        public void ReplacesBinaryByNativeType() {
            var model = root.AddClass(new CodeClass {
                Name = "model",
                ClassKind = CodeClassKind.Model
            }).First();
            var method = model.AddMethod(new CodeMethod {
                Name = "method",
                ReturnType = new CodeType {
                   Name = "binary"
                },
            }).First();
            ILanguageRefiner.Refine(new GenerationConfiguration { Language = GenerationLanguage.Java }, root);
            Assert.NotEmpty(model.StartBlock.Usings);
            Assert.NotEqual("binary", method.ReturnType.Name);
        }
        [Fact]
        public void ReplacesIndexersByMethodsWithParameter() {
            var model = root.AddClass(new CodeClass {
                Name = "model",
                ClassKind = CodeClassKind.Model
            }).First();
            var requestBuilder = root.AddClass(new CodeClass {
                Name = "requestBuilder",
                ClassKind = CodeClassKind.Model
            }).First();
            requestBuilder.AddProperty(new CodeProperty {
                Name = "pathSegment",
                DefaultValue = "path",
                Type = new CodeType {
                    Name = "string",
                }
            });
            requestBuilder.SetIndexer(new CodeIndexer {
                Name = "idx",
                ReturnType = new CodeType {
                    Name = "model",
                    TypeDefinition = model,
                },
            });
            var collectionRequestBuilder = root.AddClass(new CodeClass {
                Name = "CollectionRequestBUilder",
            }).First();
            collectionRequestBuilder.AddProperty(new CodeProperty {
                Name = "collection",
                Type = new CodeType {
                    Name = "requestBuilder",
                    TypeDefinition = requestBuilder,
                },
            });
            ILanguageRefiner.Refine(new GenerationConfiguration { Language = GenerationLanguage.Java }, root);
            Assert.Single(requestBuilder.Properties);
            Assert.Empty(requestBuilder.GetChildElements(true).OfType<CodeIndexer>());
            Assert.Single(collectionRequestBuilder.Methods.Where(x => x.IsOfKind(CodeMethodKind.IndexerBackwardCompatibility)));
            Assert.Single(collectionRequestBuilder.Properties);
        }
        [Fact]
        public void AddsInnerClasses() {
            var model = root.AddClass(new CodeClass {
                Name = "model",
                ClassKind = CodeClassKind.Model
            }).First();
            var method = model.AddMethod(new CodeMethod {
                Name = "method1",
                ReturnType = new CodeType {
                    Name = "string",
                    IsExternal = true
                }
            }).First();
            var parameter = new CodeParameter {
                Name = "param1",
                ParameterKind = CodeParameterKind.QueryParameter,
                Type = new CodeType {
                    Name = "SomeCustomType",
                    ActionOf = true,
                    TypeDefinition = new CodeClass {
                        Name = "SomeCustomType"
                    }
                }
            };
            method.AddParameter(parameter);
            ILanguageRefiner.Refine(new GenerationConfiguration { Language = GenerationLanguage.Java }, root);
            Assert.Equal(2, model.GetChildElements(true).Count());
        }
        #endregion
        #region JavaLanguageRefinerTests
        [Fact]
        public void AddsEnumSetImport() {
            var model = root.AddClass(new CodeClass {
                Name = "model",
                ClassKind = CodeClassKind.Model
            }).First();
            model.AddProperty(new CodeProperty{
                Name = "prop1",
                Type = new CodeType {
                    Name = "SomeEnum",
                    TypeDefinition = new CodeEnum {
                        Name = "SomeEnum",
                        Flags = true,
                        Parent = root,
                    }
                }
            });
            ILanguageRefiner.Refine(new GenerationConfiguration { Language = GenerationLanguage.Java }, root);
            Assert.NotEmpty((model.StartBlock as CodeClass.Declaration).Usings.Where(x => "EnumSet".Equals(x.Name)));
        }
        [Fact]
        public void CorrectsCoreType() {
            const string requestAdapterDefaultName = "IRequestAdapter";
            const string factoryDefaultName = "ISerializationWriterFactory";
            const string deserializeDefaultName = "IDictionary<string, Action<Model, IParseNode>>";
            const string dateTimeOffsetDefaultName = "DateTimeOffset";
            const string addiationalDataDefaultName = "new Dictionary<string, object>()";
            var model = root.AddClass(new CodeClass {
                Name = "model",
                ClassKind = CodeClassKind.Model
            }).First();
            model.AddProperty(new () {
                Name = "core",
                PropertyKind = CodePropertyKind.RequestAdapter,
                Type = new CodeType {
                    Name = requestAdapterDefaultName
                }
            }, new () {
                Name = "someDate",
                PropertyKind = CodePropertyKind.Custom,
                Type = new CodeType {
                    Name = dateTimeOffsetDefaultName,
                }
            }, new () {
                Name = "additionalData",
                PropertyKind = CodePropertyKind.AdditionalData,
                Type = new CodeType {
                    Name = addiationalDataDefaultName
                }
            });
            const string handlerDefaultName = "IResponseHandler";
            const string headersDefaultName = "IDictionary<string, string>";
            var executorMethod = model.AddMethod(new CodeMethod {
                Name = "executor",
                MethodKind = CodeMethodKind.RequestExecutor,
                ReturnType = new CodeType {
                    Name = "string"
                }
            }, new () {
                Name = "deserializeFields",
                ReturnType = new CodeType {
                    Name = deserializeDefaultName,
                },
                MethodKind = CodeMethodKind.Deserializer
            }).First();
            executorMethod.AddParameter(new () {
                Name = "handler",
                ParameterKind = CodeParameterKind.ResponseHandler,
                Type = new CodeType {
                    Name = handlerDefaultName,
                }
            }, new () {
                Name = "headers",
                ParameterKind = CodeParameterKind.Headers,
                Type = new CodeType() {
                    Name = headersDefaultName
                }
            });
            const string serializerDefaultName = "ISerializationWriter";
            var serializationMethod = model.AddMethod(new CodeMethod {
                Name = "seriailization",
                MethodKind = CodeMethodKind.Serializer,
                ReturnType = new CodeType {
                    Name = "string"
                }
            }).First();
            serializationMethod.AddParameter(new CodeParameter {
                Name = "handler",
                ParameterKind = CodeParameterKind.Serializer,
                Type = new CodeType {
                    Name = serializerDefaultName,
                }
            });
            ILanguageRefiner.Refine(new GenerationConfiguration { Language = GenerationLanguage.Java }, root);
            Assert.Empty(model.Properties.Where(x => requestAdapterDefaultName.Equals(x.Type.Name)));
            Assert.Empty(model.Properties.Where(x => factoryDefaultName.Equals(x.Type.Name)));
            Assert.Empty(model.Properties.Where(x => dateTimeOffsetDefaultName.Equals(x.Type.Name)));
            Assert.Empty(model.Properties.Where(x => addiationalDataDefaultName.Equals(x.Type.Name)));
            Assert.Empty(model.Methods.Where(x => deserializeDefaultName.Equals(x.ReturnType.Name)));
            Assert.Empty(model.Methods.SelectMany(x => x.Parameters).Where(x => handlerDefaultName.Equals(x.Type.Name)));
            Assert.Empty(model.Methods.SelectMany(x => x.Parameters).Where(x => headersDefaultName.Equals(x.Type.Name)));
            Assert.Empty(model.Methods.SelectMany(x => x.Parameters).Where(x => serializerDefaultName.Equals(x.Type.Name)));
        }
        [Fact]
        public void AddsMethodsOverloads() {
            var builder = root.AddClass(new CodeClass {
                Name = "model",
                ClassKind = CodeClassKind.RequestBuilder
            }).First();
            var executor = builder.AddMethod(new CodeMethod {
                Name = "executor",
                MethodKind = CodeMethodKind.RequestExecutor,
                ReturnType = new CodeType {
                    Name = "string"
                }
            }).First();
            executor.AddParameter(new CodeParameter {
                Name = "handler",
                ParameterKind = CodeParameterKind.ResponseHandler,
                Type = new CodeType {
                    Name = "string"
                }
            });
            executor.AddParameter(new CodeParameter {
                Name = "headers",
                ParameterKind = CodeParameterKind.Headers,
                Type = new CodeType {
                    Name = "string"
                }
            });
            executor.AddParameter(new CodeParameter {
                Name = "query",
                ParameterKind = CodeParameterKind.QueryParameter,
                Type = new CodeType {
                    Name = "string"
                }
            });
            executor.AddParameter(new CodeParameter {
                Name = "body",
                ParameterKind = CodeParameterKind.RequestBody,
                Type = new CodeType {
                    Name = "string"
                }
            });
            executor.AddParameter(new CodeParameter {
                Name = "options",
                ParameterKind = CodeParameterKind.Options,
                Type = new CodeType {
                    Name = "string"
                }
            });
            var generator = builder.AddMethod(new CodeMethod {
                Name = "generator",
                MethodKind = CodeMethodKind.RequestGenerator,
                ReturnType = new CodeType {
                    Name = "string"
                }
            }).First();
            generator.AddParameter(executor.Parameters.Where(x => !x.IsOfKind(CodeParameterKind.ResponseHandler)).ToArray());
            ILanguageRefiner.Refine(new GenerationConfiguration { Language = GenerationLanguage.Java }, root);
            var childMethods = builder.Methods;
            Assert.Contains(childMethods, x => x.IsOverload && x.IsOfKind(CodeMethodKind.RequestExecutor) && x.Parameters.Count() == 1);//only the body
            Assert.Contains(childMethods, x => x.IsOverload && x.IsOfKind(CodeMethodKind.RequestGenerator) && x.Parameters.Count() == 1);//only the body
            Assert.Contains(childMethods, x => x.IsOverload && x.IsOfKind(CodeMethodKind.RequestExecutor) && x.Parameters.Count() == 2);// body + query params
            Assert.Contains(childMethods, x => x.IsOverload && x.IsOfKind(CodeMethodKind.RequestGenerator) && x.Parameters.Count() == 2);// body + query params
            Assert.Contains(childMethods, x => x.IsOverload && x.IsOfKind(CodeMethodKind.RequestExecutor) && x.Parameters.Count() == 3);// body + query params + headers
            Assert.Contains(childMethods, x => x.IsOverload && x.IsOfKind(CodeMethodKind.RequestGenerator) && x.Parameters.Count() == 3);// body + query params + headers
            Assert.Contains(childMethods, x => x.IsOverload && x.IsOfKind(CodeMethodKind.RequestExecutor) && x.Parameters.Count() == 4);// body + query params + headers + options
            Assert.Contains(childMethods, x => !x.IsOverload && x.IsOfKind(CodeMethodKind.RequestGenerator) && x.Parameters.Count() == 4);// body + query params + headers + options
            Assert.Contains(childMethods, x => !x.IsOverload && x.IsOfKind(CodeMethodKind.RequestExecutor) && x.Parameters.Count() == 5);// body + query params + headers + options + response handler
            Assert.Equal(9, childMethods.Count());
            Assert.Equal(7, childMethods.Count(x => x.IsOverload));
        }
        #endregion
    }
}
